@@ -67,6 +67,23 @@ class libvirt_open(object):
         self._conn.close()
 
 
+def power_cycle(wrapped):
+    def wrapper(self, identity, *args, **kwargs):
+        power_state = self.get_power_state(identity)
+
+        if power_state == 'On':
+            self.set_power_state(identity, 'ForceOff')
+
+        try:
+            return wrapped(self, identity, *args, **kwargs)
+
+        finally:
+            if power_state == 'On':
+                self.set_power_state(identity, power_state)
+
+    return wrapper
+
+
 class LibvirtDriver(AbstractSystemsDriver):
     """Libvirt driver"""
 
@@ -81,17 +98,31 @@ class LibvirtDriver(AbstractSystemsDriver):
 
     BOOT_DEVICE_MAP_REV = {v: k for k, v in BOOT_DEVICE_MAP.items()}
 
+    DISK_DEVICE_MAP = {
+        constants.DEVICE_TYPE_HDD: 'disk',
+        constants.DEVICE_TYPE_CD: 'cdrom',
+        constants.DEVICE_TYPE_FLOPPY: 'floppy'
+    }
+
+    DISK_DEVICE_MAP_REV = {v: k for k, v in DISK_DEVICE_MAP.items()}
+
+    INTERFACE_MAP = {
+        constants.DEVICE_TYPE_PXE: 'network',
+    }
+
+    INTERFACE_MAP_REV = {v: k for k, v in INTERFACE_MAP.items()}
+
     LIBVIRT_URI = 'qemu:///system'
 
     BOOT_MODE_MAP = {
         'Legacy': 'rom',
-        'Uefi': 'pflash'
+        'UEFI': 'pflash'
     }
 
     BOOT_MODE_MAP_REV = {v: k for k, v in BOOT_MODE_MAP.items()}
 
     BOOT_LOADER_MAP = {
-        'Uefi': {
+        'UEFI': {
             'x86_64': '/usr/share/OVMF/OVMF_CODE.fd',
             'aarch64': '/usr/share/AAVMF/AAVMF_CODE.fd'
         },
@@ -109,10 +140,10 @@ class LibvirtDriver(AbstractSystemsDriver):
 
     DEVICE_TYPE_MAP_REV = {v: k for k, v in DEVICE_TYPE_MAP.items()}
 
-    # target device, controller ID, controller unit number
+    # target device, controller ID
     DEVICE_TARGET_MAP = {
-        constants.DEVICE_TYPE_FLOPPY: ('fda', 'fdc', '0'),
-        constants.DEVICE_TYPE_CD: ('hdc', 'ide', '1')
+        constants.DEVICE_TYPE_FLOPPY: ('fda', 'fdc'),
+        constants.DEVICE_TYPE_CD: ('hdc', 'ide')
     }
 
     DEFAULT_BIOS_ATTRIBUTES = {"BootMode": "Uefi",
@@ -225,7 +256,6 @@ class LibvirtDriver(AbstractSystemsDriver):
 
         :param identity: libvirt domain name or ID
         :param state: string literal requesting power state transition.
-            If not specified, current system power state is returned.
             Valid values  are: *On*, *ForceOn*, *ForceOff*, *GracefulShutdown*,
             *GracefulRestart*, *ForceRestart*, *Nmi*.
 
@@ -262,6 +292,10 @@ class LibvirtDriver(AbstractSystemsDriver):
     def get_boot_device(self, identity):
         """Get computer system boot device name
 
+        First try to get boot device from bootloader configuration.. If it's
+        not present, proceed towards gathering boot order information from
+        per-device boot configuration, then pick the lowest ordered device.
+
         :param identity: libvirt domain name or ID
 
         :returns: boot device name as `str` or `None` if device name
@@ -271,21 +305,79 @@ class LibvirtDriver(AbstractSystemsDriver):
 
         tree = ET.fromstring(domain.XMLDesc())
 
+        # Try boot configuration in the bootloader
+
         boot_element = tree.find('.//boot')
         if boot_element is not None:
-            boot_source_target = (
-                self.BOOT_DEVICE_MAP_REV.get(boot_element.get('dev'))
-            )
+            dev_attr = boot_element.get('dev')
+            if dev_attr is not None:
+                boot_source_target = self.BOOT_DEVICE_MAP_REV.get(dev_attr)
+                if boot_source_target:
+                    return boot_source_target
 
-            return boot_source_target
+        min_order = boot_source_target = None
 
+        # If bootloader config is not present, try per-device boot elements
+
+        devices_element = tree.find('devices')
+
+        if devices_element is not None:
+
+            for disk_element in devices_element.findall('disk'):
+                boot_element = disk_element.find('boot')
+                if boot_element is None:
+                    continue
+
+                order = boot_element.get('order')
+                if not order:
+                    continue
+
+                order = int(order)
+                if min_order is not None and order >= min_order:
+                    continue
+
+                device_attr = disk_element.get('device')
+                if device_attr is None:
+                    continue
+
+                boot_source_target = self.DISK_DEVICE_MAP_REV.get(
+                    device_attr)
+
+                if boot_source_target:
+                    min_order = order
+
+            for interface_element in devices_element.findall('interface'):
+                boot_element = interface_element.find('boot')
+                if boot_element is None:
+                    continue
+
+                order = boot_element.get('order')
+                if not order:
+                    continue
+
+                order = int(order)
+                if min_order is not None and order >= min_order:
+                    continue
+
+                boot_source_target = self.INTERFACE_MAP_REV.get('network')
+
+                if boot_source_target:
+                    min_order = order
+
+        return boot_source_target
+
+    @power_cycle
     def set_boot_device(self, identity, boot_source):
         """Get/Set computer system boot device name
 
+        First remove all boot device configuration from bootloader because
+        that's legacy with libvirt. Then remove possible boot configuration
+        in the per-device settings. Finally, make the desired boot device
+        the only bootable by means of per-device configuration boot option.
+
         :param identity: libvirt domain name or ID
-        :param boot_source: optional string literal requesting boot device
-            change on the system. If not specified, current boot device is
-            returned. Valid values are: *Pxe*, *Hdd*, *Cd*.
+        :param boot_source: string literal requesting boot device
+            change on the system. Valid values are: *Pxe*, *Hdd*, *Cd*.
 
         :raises: `error.FishyError` if boot device can't be set
         """
@@ -294,23 +386,65 @@ class LibvirtDriver(AbstractSystemsDriver):
         # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
         tree = ET.fromstring(domain.XMLDesc())
 
-        try:
-            target = self.BOOT_DEVICE_MAP[boot_source]
-
-        except KeyError:
-            msg = ('Unknown power state requested: '
-                   '%(boot_source)s' % {'boot_source': boot_source})
-
-            raise error.FishyError(msg)
+        # Remove bootloader configuration
 
         for os_element in tree.findall('os'):
-            # Remove all "boot" elements
             for boot_element in os_element.findall('boot'):
                 os_element.remove(boot_element)
 
-            # Add a new boot element with the request boot device
-            boot_element = ET.SubElement(os_element, 'boot')
-            boot_element.set('dev', target)
+        target = self.DISK_DEVICE_MAP.get(boot_source)
+
+        # Process per-device boot configuration
+
+        devices_element = tree.find('devices')
+        if devices_element is None:
+            msg = ('Incomplete libvirt domain configuration - <devices> '
+                   'element is missing in domain '
+                   '%(uuid)s' % {'uuid': domain.UUIDString()})
+
+            raise error.FishyError(msg)
+
+        target_device_elements = []
+
+        # Remove per-disk boot configuration
+
+        for disk_element in devices_element.findall('disk'):
+
+            device_attr = disk_element.get('device')
+            if device_attr is None:
+                continue
+
+            # NOTE(etingof): multiple devices of the same type not supported
+            if device_attr == target:
+                target_device_elements.append(disk_element)
+
+            for boot_element in disk_element.findall('boot'):
+                disk_element.remove(boot_element)
+
+        target = self.INTERFACE_MAP.get(boot_source)
+
+        # Remove per-interface boot configuration
+
+        for interface_element in devices_element.findall('interface'):
+
+            if target == 'network':
+                target_device_elements.append(interface_element)
+
+            for boot_element in interface_element.findall('boot'):
+                interface_element.remove(boot_element)
+
+        if not target_device_elements:
+            msg = ('Target libvirt device %(target)s does not exist in domain '
+                   '%(uuid)s' % {'target': boot_source,
+                                 'uuid': domain.UUIDString()})
+
+            raise error.FishyError(msg)
+
+        # NOTE(etingof): Make all chosen devices bootable (important for NICs)
+
+        for order, target_device_element in enumerate(target_device_elements):
+            boot_element = ET.SubElement(target_device_element, 'boot')
+            boot_element.set('order', str(order + 1))
 
         try:
             with libvirt_open(self._uri) as conn:
@@ -325,7 +459,7 @@ class LibvirtDriver(AbstractSystemsDriver):
     def get_boot_mode(self, identity):
         """Get computer system boot mode.
 
-        :returns: either *Uefi* or *Legacy* as `str` or `None` if
+        :returns: either *UEFI* or *Legacy* as `str` or `None` if
             current boot mode can't be determined
         """
         domain = self._get_domain(identity, readonly=True)
@@ -342,12 +476,12 @@ class LibvirtDriver(AbstractSystemsDriver):
 
             return boot_mode
 
+    @power_cycle
     def set_boot_mode(self, identity, boot_mode):
         """Set computer system boot mode.
 
-        :param boot_mode: optional string literal requesting boot mode
-            change on the system. If not specified, current boot mode is
-            returned. Valid values are: *Uefi*, *Legacy*.
+        :param boot_mode: string literal requesting boot mode
+            change on the system. Valid values are: *UEFI*, *Legacy*.
 
         :raises: `error.FishyError` if boot mode can't be set
         """
@@ -587,6 +721,7 @@ class LibvirtDriver(AbstractSystemsDriver):
         """
         return self._process_bios(identity)
 
+    @power_cycle
     def set_bios(self, identity, attributes):
         """Update BIOS attributes
 
@@ -609,6 +744,7 @@ class LibvirtDriver(AbstractSystemsDriver):
         self._process_bios(identity, bios_attributes,
                            update_existing_attributes=True)
 
+    @power_cycle
     def reset_bios(self, identity):
         """Reset BIOS attributes to default
 
@@ -670,7 +806,7 @@ class LibvirtDriver(AbstractSystemsDriver):
             inserted = (self.get_boot_device(identity) ==
                         constants.DEVICE_TYPE_CD)
             if inserted:
-                inserted = self.get_boot_mode(identity) == 'Uefi'
+                inserted = self.get_boot_mode(identity) == 'UEFI'
 
             return boot_image, read_only, inserted
 
@@ -748,28 +884,60 @@ class LibvirtDriver(AbstractSystemsDriver):
                 raise error.FishyError(
                     'Unknown device %s at %s' % (device, identity))
 
+            tgt_dev, tgt_bus = self.DEVICE_TARGET_MAP[device]
+
+            # Enumerate existing disks to find a free unit on the bus
+
+            free_units = {i for i in range(100)}
+
+            disk_elements = device_element.findall('disk')
+
+            for disk_element in disk_elements:
+                target_element = disk_element.find('target')
+                if target_element is None:
+                    continue
+
+                bus_type = target_element.attrib.get('bus')
+                if bus_type != tgt_bus:
+                    continue
+
+                address_element = disk_element.find('address')
+                if address_element is None:
+                    continue
+
+                unit_num = address_element.attrib.get('unit')
+                if unit_num is None:
+                    continue
+
+                if unit_num in free_units:
+                    free_units.remove(unit_num)
+
+            if not free_units:
+                msg = ('No free %(bus)s bus unit found in the libvirt domain '
+                       '"%(identity)s" configuration' % {'identity': identity,
+                                                         'bus': tgt_bus})
+                raise error.FishyError(msg)
+
             # Add disk element pointing to the boot image
 
             disk_element = ET.SubElement(device_element, 'disk')
             disk_element.set('type', 'file')
             disk_element.set('device', lv_device)
 
-            tgt_dev, tgt_bus, tgt_unit = self.DEVICE_TARGET_MAP[device]
-
             target_element = ET.SubElement(disk_element, 'target')
             target_element.set('dev', tgt_dev)
             target_element.set('bus', tgt_bus)
 
-            driver_element = ET.SubElement(disk_element, 'driver')
-            driver_element.set('name', 'qemu')
-            driver_element.set('type', 'raw')
-
             address_element = ET.SubElement(disk_element, 'address')
             address_element.set('type', 'drive')
             address_element.set('controller', '0')
             address_element.set('bus', '0')
             address_element.set('target', '0')
-            address_element.set('unit', tgt_unit)
+            address_element.set('unit', '%s' % min(free_units))
+
+            driver_element = ET.SubElement(disk_element, 'driver')
+            driver_element.set('name', 'qemu')
+            driver_element.set('type', 'raw')
 
             source_element = ET.SubElement(disk_element, 'source')
             source_element.set('file', image_path)
@@ -781,12 +949,6 @@ class LibvirtDriver(AbstractSystemsDriver):
 
         identity = domain.UUIDString()
 
-        device_element = domain_tree.find('devices')
-        if device_element is None:
-            msg = ('Missing "devices" tag in the libvirt domain '
-                   '"%(identity)s" configuration' % {'identity': identity})
-            raise error.FishyError(msg)
-
         try:
             lv_device = self.BOOT_DEVICE_MAP[device]
 
@@ -800,7 +962,7 @@ class LibvirtDriver(AbstractSystemsDriver):
                    '"%(identity)s" configuration' % {'identity': identity})
             raise error.FishyError(msg)
 
-        # First, remove all existing devices
+        # Remove all existing devices
         disk_elements = device_element.findall('disk')
 
         for disk_element in disk_elements:
@@ -808,6 +970,7 @@ class LibvirtDriver(AbstractSystemsDriver):
             if dev_type == lv_device:
                 device_element.remove(disk_element)
 
+    @power_cycle
     def set_boot_image(self, identity, device, boot_image=None,
                        write_protected=True):
         """Set backend VM boot image
@@ -827,20 +990,30 @@ class LibvirtDriver(AbstractSystemsDriver):
 
         self._remove_boot_images(domain, domain_tree, device)
 
+        boot_device = None
+
         if boot_image:
+            self._add_boot_image(domain, domain_tree, device,
+                                 boot_image, write_protected)
+
+            boot_device = self.get_boot_device(identity)
+
+        with libvirt_open(self._uri) as conn:
+            xml = ET.tostring(domain_tree)
 
             try:
-                self._add_boot_image(domain, domain_tree, device,
-                                     boot_image, write_protected)
+                conn.defineXML(xml.decode('utf-8'))
+
+            except Exception as e:
+                self._logger.error('Rejected libvirt domain XML is %s', xml)
 
-            except libvirt.libvirtError as e:
                 msg = ('Error changing boot image at libvirt URI "%(uri)s": '
                        '%(error)s' % {'uri': self._uri, 'error': e})
 
                 raise error.FishyError(msg)
 
-        with libvirt_open(self._uri) as conn:
-            conn.defineXML(ET.tostring(domain_tree).decode('utf-8'))
+        if device == boot_device:
+            self.set_boot_device(identity, boot_device)
 
     def _find_device_by_path(self, vol_path):
         """Get device attributes using path
