@@ -161,9 +161,50 @@ class LibvirtDriverTestCase(base.BaseTestCase):
 
         domain_mock.injectNMI.assert_called_once_with()
 
+    @mock.patch('libvirt.open', autospec=True)
+    def test_power_cycle_when_off(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_os.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        with mock.patch.object(
+                self.test_driver, 'get_power_state') as gps_mock:
+            with mock.patch.object(
+                    self.test_driver, 'set_power_state') as sps_mock:
+                gps_mock.return_value = 'Off'
+                self.test_driver.set_boot_device(self.uuid, 'Cd')
+
+        self.assertTrue(gps_mock.called)
+        self.assertFalse(sps_mock.called)
+
+    @mock.patch('libvirt.open', autospec=True)
+    def test_power_cycle_when_on(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_os.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        with mock.patch.object(
+                self.test_driver, 'get_power_state') as gps_mock:
+            with mock.patch.object(
+                    self.test_driver, 'set_power_state') as sps_mock:
+                gps_mock.return_value = 'On'
+                self.test_driver.set_boot_device(self.uuid, 'Cd')
+
+        self.assertTrue(gps_mock.called)
+        self.assertTrue(sps_mock.called)
+
     @mock.patch('libvirt.openReadOnly', autospec=True)
-    def test_get_boot_device(self, libvirt_mock):
-        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
+    def test_get_boot_device_os(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_os.xml', 'r') as f:
             data = f.read()
 
         conn_mock = libvirt_mock.return_value
@@ -175,18 +216,95 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         self.assertEqual('Cd', boot_device)
 
     @mock.patch('libvirt.open', autospec=True)
-    def test_set_boot_device(self, libvirt_mock):
-        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
+    def test_set_boot_device_os(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_os.xml', 'r') as f:
             data = f.read()
 
         conn_mock = libvirt_mock.return_value
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.test_driver.set_boot_device(self.uuid, 'Hdd')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_boot_device(self.uuid, 'Hdd')
 
         conn_mock.defineXML.assert_called_once_with(mock.ANY)
 
+    @mock.patch('libvirt.openReadOnly', autospec=True)
+    def test_get_boot_device_disk(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_disk.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        boot_device = self.test_driver.get_boot_device(self.uuid)
+
+        self.assertEqual('Cd', boot_device)
+
+    @mock.patch('libvirt.open', autospec=True)
+    def test_set_boot_device_disk(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_disk.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_boot_device(self.uuid, 'Hdd')
+
+        conn_mock.defineXML.assert_called_once_with(mock.ANY)
+
+        expected = '<disk device="disk" type="file">\n      <source ' \
+                   'file="/home/user/fedora.img" />\n      <target ' \
+                   'dev="hda" />\n    <boot order="1" /></disk>\n'
+
+        self.assertIn(expected, conn_mock.defineXML.call_args[0][0])
+
+    @mock.patch('libvirt.openReadOnly', autospec=True)
+    def test_get_boot_device_network(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_network.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        boot_device = self.test_driver.get_boot_device(self.uuid)
+
+        self.assertEqual('Pxe', boot_device)
+
+    @mock.patch('libvirt.open', autospec=True)
+    def test_set_boot_device_network(self, libvirt_mock):
+        with open('sushy_tools/tests/unit/emulator/'
+                  'domain_boot_network.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_boot_device(self.uuid, 'Pxe')
+
+        conn_mock.defineXML.assert_called_once_with(mock.ANY)
+
+        expected = '<interface type="direct">\n      <mac address=' \
+                   '"52:54:00:da:ac:54" />\n      <source dev="tap-' \
+                   'node-2i1" mode="vepa" />\n      <model type="vir' \
+                   'tio" />\n      <address bus="0x01" domain="0x' \
+                   '0000" function="0x0" slot="0x01" type="pci" />'
+
+        self.assertIn(expected, conn_mock.defineXML.call_args[0][0])
+
     @mock.patch('libvirt.openReadOnly', autospec=True)
     def test_get_boot_mode(self, libvirt_mock):
         with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
@@ -210,7 +328,9 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.test_driver.set_boot_mode(self.uuid, 'Uefi')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_boot_mode(self.uuid, 'UEFI')
 
         conn_mock = libvirt_rw_mock.return_value
         conn_mock.defineXML.assert_called_once_with(mock.ANY)
@@ -225,7 +345,9 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.test_driver.set_boot_mode(self.uuid, 'Uefi')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_boot_mode(self.uuid, 'UEFI')
 
         conn_mock = libvirt_rw_mock.return_value
         xml_document = conn_mock.defineXML.call_args[0][0]
@@ -251,9 +373,11 @@ class LibvirtDriverTestCase(base.BaseTestCase):
 
         with mock.patch.dict(
                 self.test_driver.KNOWN_BOOT_LOADERS, {}, clear=True):
-            self.assertRaises(
-                error.FishyError, self.test_driver.set_boot_mode,
-                self.uuid, 'Uefi')
+            with mock.patch.object(
+                    self.test_driver, 'get_power_state', return_value='Off'):
+                self.assertRaises(
+                    error.FishyError, self.test_driver.set_boot_mode,
+                    self.uuid, 'Uefi')
 
     @mock.patch('libvirt.open', autospec=True)
     @mock.patch('libvirt.openReadOnly', autospec=True)
@@ -271,9 +395,11 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.assertRaises(
-            error.FishyError, self.test_driver.set_boot_mode,
-            self.uuid, 'Uefi')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.assertRaises(
+                error.FishyError, self.test_driver.set_boot_mode,
+                self.uuid, 'Uefi')
 
     @mock.patch('libvirt.open', autospec=True)
     @mock.patch('libvirt.openReadOnly', autospec=True)
@@ -291,9 +417,11 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.assertRaises(
-            error.FishyError, self.test_driver.set_boot_mode,
-            self.uuid, 'Uefi')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.assertRaises(
+                error.FishyError, self.test_driver.set_boot_mode,
+                self.uuid, 'Uefi')
 
     @mock.patch('libvirt.open', autospec=True)
     @mock.patch('libvirt.openReadOnly', autospec=True)
@@ -311,9 +439,11 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.assertRaises(
-            error.FishyError, self.test_driver.set_boot_mode,
-            self.uuid, 'Uefi')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.assertRaises(
+                error.FishyError, self.test_driver.set_boot_mode,
+                self.uuid, 'Uefi')
 
     @mock.patch('libvirt.open', autospec=True)
     @mock.patch('libvirt.openReadOnly', autospec=True)
@@ -332,7 +462,9 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = data
 
-        self.test_driver.set_boot_mode(self.uuid, 'Uefi')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_boot_mode(self.uuid, 'UEFI')
 
         conn_mock = libvirt_rw_mock.return_value
         xml_document = conn_mock.defineXML.call_args[0][0]
@@ -378,7 +510,13 @@ class LibvirtDriverTestCase(base.BaseTestCase):
 
         pool_mock.XMLDesc.return_value = data
 
-        self.test_driver.set_boot_image(self.uuid, 'Cd', '/tmp/image.iso')
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            with mock.patch.object(
+                    self.test_driver, 'get_boot_device', return_value=None):
+
+                self.test_driver.set_boot_image(
+                    self.uuid, 'Cd', '/tmp/image.iso')
 
         conn_mock = libvirt_rw_mock.return_value
         pool_mock.listAllVolumes.assert_called_once_with()
@@ -390,6 +528,39 @@ class LibvirtDriverTestCase(base.BaseTestCase):
 
         conn_mock.defineXML.assert_called_once_with(mock.ANY)
 
+    @mock.patch('libvirt.open', autospec=True)
+    @mock.patch('libvirt.openReadOnly', autospec=True)
+    @mock.patch(
+        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
+        '.get_power_state', new=mock.MagicMock(return_value='Off'))
+    @mock.patch(
+        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
+        '.get_boot_device', return_value='Cd')
+    @mock.patch(
+        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
+        '.set_boot_device')
+    @mock.patch(
+        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
+        '._add_boot_image', new=mock.MagicMock())
+    @mock.patch(
+        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
+        '._remove_boot_images', new=mock.MagicMock())
+    def test_set_boot_image_restore_boot_device(
+            self, sbd_mock, gbd_mock, libvirt_mock, libvirt_rw_mock):
+
+        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
+            data = f.read()
+
+        conn_mock = libvirt_rw_mock.return_value
+        domain_mock = conn_mock.lookupByUUID.return_value
+        domain_mock.XMLDesc.return_value = data
+
+        self.test_driver.set_boot_image(
+            self.uuid, 'Cd', '/tmp/image.iso')
+
+        gbd_mock.assert_called_once_with(self.uuid)
+        sbd_mock.assert_called_once_with(self.uuid, 'Cd')
+
     @mock.patch('libvirt.openReadOnly', autospec=True)
     def test_get_total_memory(self, libvirt_mock):
         conn_mock = libvirt_mock.return_value
@@ -452,9 +623,12 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = domain_xml
 
-        self.test_driver.set_bios(self.uuid,
-                                  {"BootMode": "Uefi",
-                                   "ProcTurboMode": "Enabled"})
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.set_bios(
+                self.uuid, {"BootMode": "Uefi",
+                            "ProcTurboMode": "Enabled"})
+
         conn_mock.defineXML.assert_called_once_with(mock.ANY)
 
     @mock.patch('libvirt.open', autospec=True)
@@ -466,7 +640,10 @@ class LibvirtDriverTestCase(base.BaseTestCase):
         domain_mock = conn_mock.lookupByUUID.return_value
         domain_mock.XMLDesc.return_value = domain_xml
 
-        self.test_driver.reset_bios(self.uuid)
+        with mock.patch.object(
+                self.test_driver, 'get_power_state', return_value='Off'):
+            self.test_driver.reset_bios(self.uuid)
+
         conn_mock.defineXML.assert_called_once_with(mock.ANY)
 
     def test__process_bios_attributes_get_default(self):
